package ch.usi.inf.sape.tracer.analyzer;

import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BytecodeEvent;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.CommonSubExpression;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.UntracedBytecode_Operation;
import ch.usi.inf.sape.tracer.analyzer.locations.LocalMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocationInfo;
import ch.usi.inf.sape.tracer.analyzer.locations.StackSlotMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocationInfo.Type;

/**
 * A memory location can be defined several times during its life cyclse. Each instance of this
 * class represents a single definition of particular memory location.
 * 
 * @author reza
 *
 */
public final class Binding<T extends MemoryLocation> {

	/**
	 * The memory location to which this binding applies.
	 */
	private final T memoryLocation;
	
	/**
	 * The bytecode operation that generated this binding
	 */
	private final BytecodeEvent provider;
	
	/**
	 * The value set for this binding
	 */
	private final Value value;
	
	/**
	 * is this particular binding put on top of stack already?
	 */
	private transient int occurencesOnStack;
	
//	if the value to be written (which is among the uses of an instruction location on top of stack) has a copy on the stack (binding.hasACopyOnStack() symbolicpresentation -> dest = temp(source->binding))
//	else if the instruction is overwriting a non-stack location that still has a binding on top of the stack introduce the temp as before 
	
	/**
	 * This is true if the value is already on the stack and overwritten by an instruction
	 */
	private transient boolean overwritten;
	
	/**
	 * This name is shown instead of the real name of the location in case the value of the location is overwritten
	 */
	private transient String temporaryName;
	
	/**
	 * if this binding represents a common subexpression generated by one of the derivatives of dup instruction that has another copy on the stack 
	 */
	private transient boolean isCommonSubexpression;
	
	
	private transient boolean isSubTreeEvaluated;
	
	/**
	 * This is used when we know there will be an NPE after the instrumented bytecode gets executed
	 * like baload, idiv, invokevirtual
	 */
	public final static Binding UNDEFINED_BINDING = new Binding(MemoryLocationInfo.UNDEFINED_MEMORY_LOCATION, Value.UNDEFINED, BytecodeEvent.UNTRACED_BYTECODE_OPERATION);
	
	/**
	 * This is used when we know there will be an NPE after the instrumented bytecode gets executed
	 * like baload, idiv, invokevirtual
	 */
	public final static Binding[] UNDEFINED_BINDINGS = new Binding[0];
	
	/**
	 * This is used when the instruction doesn't define any value like monitorenter, monitorexit, pop
	 */
	public final static Binding[] NO_DEF_BINDINGS = new Binding[0];
	
	/**
	 * This is used when the instruction doesn't use any value like bipush, sipush
	 */
	public final static Binding[] NO_USE_BINDINGS = new Binding[0];
	
	/**
	 * This is used when there is no binding that satisfies a query, e.g.: in getUseForDef
	 */
	public final static Binding[] NOT_AVAILABLE_BINDINGS = new Binding[0];
	
	/**
	 * This is used just for the uses/defs of the untraced bytecode operations
	 */
	public final static Binding[] UNTRACED_BINDINGS = new Binding[0];
	
	public Binding(T memoryLocation, Value value, BytecodeEvent provider) {
		this.memoryLocation = memoryLocation;
		this.value = value;
		this.provider = provider;
	}

	/**
	 * @return the value
	 */
	public Value getValue() {
		return value;
	}

	/**
	 * @return the memoryLocation
	 */
	public T getMemoryLocation() {
		return memoryLocation;
	}
	
	@Override
	public String toString() {
		return memoryLocation.toString() + ", value: "+ value+ ", Provider: "+provider.getOperationIndex()+", Time: "+ provider.getEventTime()+".";
	}

	/**
	 * @return the provider
	 */
	public BytecodeEvent getProvider() {
		return provider;
	}
	
	public boolean isAreadyOnTopOfStack(){
		return occurencesOnStack > 0 ? true : false;
	}
	
	public void removeCopyFromStack(){
		this.occurencesOnStack --;
	}
	
	public void addCopyOnStack(){
		this.occurencesOnStack ++;
	}

	/**
	 * @return the overwritten
	 */
	public boolean isOverwritten() {
		return overwritten;
	}

	/**
	 * @param overwritten the overwritten to set
	 */
	public void setOverwritten(boolean overwritten) {
		this.overwritten = overwritten;
	}

	/**
	 * @return the temporaryName
	 */
	public String getTemporaryName() {
		return temporaryName;
	}

	/**
	 * @param temporaryName the temporaryName to set
	 */
	public void setTemporaryName(String temporaryName) {
		this.temporaryName = temporaryName;
	}

	/**
	 * @return the isCommonSubexpression
	 */
	public boolean isCommonSubexpression() {
		return isCommonSubexpression;
	}

	/**
	 * @param isCommonSubexpression the isCommonSubexpression to set
	 */
	public void setAsCommonSubexpression() {
		this.isCommonSubexpression = true;
	}

	public String evaluate(StackSnapshot stackSnapshot){
		String result = "";
		if (!isSubTreeEvaluated){
			Binding[] uses = this.provider.getUses();
			for (Binding binding : uses) {
				result += binding.evaluate(stackSnapshot);
			}
			if (this.isCommonSubexpression() && stackSnapshot.contains(((CommonSubExpression)this.provider).getCopy(this)) ){  
				result += ((CommonSubExpression)this.provider).evaluate(this, stackSnapshot);
			}
			isSubTreeEvaluated = true;
		}
		
		return result;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((memoryLocation == null) ? 0 : memoryLocation.hashCode());
		result = prime * result
				+ ((provider == null) ? 0 : provider.hashCode());
		result = prime * result + ((value == null) ? 0 : value.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Binding other = (Binding) obj;
		if (memoryLocation == null) {
			if (other.memoryLocation != null)
				return false;
		} else if (!memoryLocation.equals(other.memoryLocation))
			return false;
		if (provider == null) {
			if (other.provider != null)
				return false;
		} else if (!provider.equals(other.provider))
			return false;
		if (value == null) {
			if (other.value != null)
				return false;
		} else if (!value.equals(other.value))
			return false;
		return true;
	}
	
	
	
}

