package ch.usi.inf.sape.tracer.analyzer;

import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocation;

public abstract class NonSlicedEvent implements Event{

	/**
	 * 
	 */
	private static final long serialVersionUID = -2277842982695184020L;

	/**
	 * The index of an event among the events of its
	 * context thread. This value is not necessarily 
	 * unique among events of different threads. For
	 * a global unique identifier, refer to GlobalClock.
	 * 
	 */
	protected final int operationIndex;
	
	/**
	 * The internal coding used for identifying bytecodes.
	 * Refer to Bytecodes interface for finding the defined
	 * codes.
	 */
	protected final short opCode;
	
	protected Binding<? extends MemoryLocation>[] uses;
	
	protected Binding<? extends MemoryLocation>[] defs;
	
	protected final int sourceLineNo;
	
	protected final Activation activationRecord;
	
	/**
	 * The global time of execution for this event.
	 * Refer to GlobalClock class for further detail.
	 */
	protected final long time;
	
	public NonSlicedEvent(short opCode, 
			int sourceLineNo, 
			int operationIndex, 
			Activation activationRecord) {
		this(opCode, sourceLineNo, operationIndex, activationRecord, null, null);
	}
	
	public NonSlicedEvent(short opCode, int sourceLineNo, 
			int operationIndex,
			Activation activationRecord,
			Binding<? extends MemoryLocation>[] uses,
			Binding<? extends MemoryLocation>[] defs){ 
		this.operationIndex = operationIndex;
		this.opCode = opCode;
		this.uses = uses;
		this.defs = defs;
		this.sourceLineNo = sourceLineNo;
		this.activationRecord = activationRecord;
		this.time = GlobalClock.getAndIncrementTime();
	}


	/**
	 * Temporary constructor for compatibility purposes
	 * TODO: remove this constructor and fix the code in AbstractOperation
	 */
	public NonSlicedEvent(int opCode, int sourceLine, int operationIndex){
		this.opCode = (short) opCode;
		this.sourceLineNo = sourceLine;
		this.operationIndex = operationIndex;
		this.activationRecord = null;
		this.time = GlobalClock.getAndIncrementTime();
	}
	

	@Override
	public Binding<? extends MemoryLocation>[] getUses() {
		return this.uses;
	}

	@Override
	public Binding<? extends MemoryLocation>[] getDefs() {
		return this.defs;
	}

	/**
	 * The index of an event among the events of its
	 * context thread. This value is not necessarily 
	 * unique among events of different threads. For
	 * a global unique identifier, refer to GlobalClock.
	 * 
	 */
	@Override
	public int getOperationIndex() {
		return operationIndex;
	}

	@Override
	public int getSourceCodeLineNumber() {
		return sourceLineNo;
	}

	/**
	 * The internal coding used for identifying bytecodes.
	 * Refer to Bytecodes interface for finding the defined
	 * codes.
	 */
	@Override
	public short getOperationCode() {
		return opCode;
	}

	@Override
	public Activation getActivationRecord() {
		return this.activationRecord;
	}
	
	@Override
	public String getSourceFileNameWithFullAddress() {
		return getActivationRecord().getSourceFileWithFullAdress(); 
	}

	@Override
	public String getSourceFileName() {
		return getActivationRecord().getSourceFile();
	}
	
	@Override
	public long getEventTime(){
		return time;
	}
	
	/**
	 * Control flow instructions include:
	 * 1- Conditional/unconditional jump instructions including:
	 * IFEQ, IFNE, IFLT, IFGE, IFGT, IFLE, IF_ICMPEQ,
     * IF_ICMPNE, IF_ICMPLT, IF_ICMPGE, IF_ICMPGT, IF_ICMPLE,
     * IF_ACMPEQ, IF_ACMPNE, GOTO, JSR, IFNULL or IFNONNULL
     * 
     * 2- Tableswitch
     * 3- Lookupswitch
     * 4- athrow (also PEI)
     * 5- Invocations (also PEI)
	 * @return
	 */
	public boolean isControlFlowInstruction(){
		return false;
	}
	
	public abstract SlicedEvent getSlicedOperationForSession(final int sessionId);
}
