package ch.usi.inf.sape.tracer.analyzer;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Stack;

import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BytecodeEvent;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.INVOKEARGS_Operation;
import ch.usi.inf.sape.tracer.analyzer.locations.Container;
import ch.usi.inf.sape.tracer.analyzer.locations.LocalMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocationInfo.Type;
import ch.usi.inf.sape.tracer.analyzer.locations.ObjectMetaData;
import ch.usi.inf.sape.tracer.analyzer.locations.PseudoVariableMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.ReturnValueMemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.StackSlotMemoryLocation;


/**
 * This class is an abstract representation of the JVM 
 * call stack for a thread. It imitates the behavior 
 * of the operand stack as well as the state of local 
 * variables table.
 * 
 * 
 * TODO: Exception handling
 * TODO: Keeping consistency with the real call stack upon 
 * calling/returning to/from an
 * uninstrumented method
 *
 */
public class ThreadCallStack extends Container{

	private static final long serialVersionUID = -5599862649711740317L;
	

	public static final ThreadCallStack UNTRACED_THREAD_CALL_STACK = new UntracedThreadCallStack();
	
	private final int threadId;
	
	/**
	 * All events generated by this thread
	 */
	private final List<BytecodeEvent> operations = new ArrayList<>();
	
	/**
	 *  All visited frames for this thread during the execution
	 */
	private final Set<Activation> frames = new HashSet<>();
	
	/**
	 *  Simulates the behavior of the call stack for this thread
	 */
	private final Stack<Activation> stack= new Stack<Activation>();
	
	/**
	 * Determines if the current thread should trace the 
	 * execution or just simply executed the original code.
	 */
	private boolean traceMode = true;
	
	/**
	 * A flag to indicate that an exception is thrown which is not caught yet.
	 * As soon as it is caught, this flag should be set to false.
	 * There are different points where we set this flag to false:
	 * 	1-upon entry to a new method (we explain the reason in the body of the 
	 * 	method "Analyzer.methodEntry"
	 * 	2- After returning from each invocation.
	 * 	3- Upon entry into a catch block. 
	 */
	private boolean exceptionThrown;
	
	/**
	 * When this flag is true, it means that an exception was thrown in an 
	 * untraced method and it is not caught yet.
	 */
	private boolean uncaughtExceptionInUninstrumented;

	/**
	 * When an exception is thrown by running an instance of ATHROW
	 * bytecode, this flag is set to true. This is different from 
	 * exceptionThrown, because it indicates the occurrence of an 
	 * uncaught exception in a method.
	 * 
	 * This flag is reset as soon
	 * as the the exception is caught. This means the same three events
	 * that can make exceptionThrown to false.
	 */
	private boolean throwInstructionOccured;
	
	private static final ThreadLocal<ThreadCallStack> currentThreadCallStack = new ThreadLocal<ThreadCallStack>();
	
	/**
	 * Keeps track of the of basic blocks visited during execution, used for
	 * visualization purposes
	 */
	private final List<PseudoVariableMemoryLocation> bbVarLocations = new Stack<>();
	
	/**
	 * Simulates the control dependency relationships
	 */
	private final Stack<Binding<PseudoVariableMemoryLocation>> basicBlocks = new Stack<>();
	
	/**
	 * This is used for visualization purposes. It indicates the offset of 
	 * the sub-table representing this activation in the whole table
	 */
	private int offset = -1; 			
	
	/**
	 * This field is used in control dependency analysis to distinguish 
	 * among different inovcations.
	 */
	private int nextInvocationId = -1;
	
	/**
	 * The next operation index for this thread. Each thread has its own 
	 * set of operations.
	 */
	private int nextOperationIndex = 0;
	
	/**
	 * The id of the basic block which is currently being executed
	 */
	private int currentBasicBlockId = -1;
	
	private static int nextThreadId = 0;
	
	/**
	 * We instantiate an object of this class only through
	 * the barrier method createNewCallStackBarrier().
	 * Never instantiate directly!
	 */
	private ThreadCallStack(int tid){
		this.threadId = tid; 
	}
	
	/**
	 * We use this method to perform all post initialization
	 * actions after creating an instance of a thread call
	 * stack. This is used for avoiding from passing a newly
	 * created obj reference to other methods.
	 * 
	 * @return
	 */
	private static ThreadCallStack createNewCallStackBarrier(){
		ThreadCallStack tcs = new ThreadCallStack(nextThreadId ++);
		RuntimeEnv.getInstance().addThreadCallStack(tcs);
		return tcs;
	}
	
	public static ThreadCallStack getCurrentThreadCallStack(){
		ThreadCallStack tcs = currentThreadCallStack.get();
		if (tcs == null){
			tcs = createNewCallStackBarrier();
			ThreadCallStack.setCurrentThreadCallStack(tcs);
		}
		
		return tcs;
	}
	
	public static void setCurrentThreadCallStack(ThreadCallStack tcs){
		currentThreadCallStack.set(tcs);
	}

	public Activation createAndPushFrame(final String methodName, 
											final String desc, 
											final String ownerClass, 
											final int argsCount, 
											final boolean isStatic, 
											final INVOKEARGS_Operation provider) {
		Activation frame = new Activation(methodName, desc, ownerClass, argsCount, isStatic, provider, ThreadCallStack.getCurrentThreadCallStack());
		pushFrame(frame);
		return frame;
	}
	
	public void setLastInvokationFrame(boolean callerInstrumented){
		if (stack.size() < 2)
			return;
		
		if (!callerInstrumented)
			return;
		
		Activation calleeFrame = popFrame();
		Activation callerFrame = peekTopMostFrame();
		callerFrame.setLastInvokationFrame(calleeFrame);
		pushFrame(calleeFrame);
	}
	
	private void pushFrame(Activation frame){
		this.stack.push(frame);
		this.frames.add(frame);
		
		int currentOperationId = operations.size();
			 
		frame.updateStartTime(currentOperationId);
	}

	public Activation popFrame(){
		Activation toBePoppedframe = stack.pop();
		int currentOperationId = operations.size();
		toBePoppedframe.updateEndTime(currentOperationId);
		return toBePoppedframe;
	}

	public Activation peekTopMostFrame(){
		return stack.peek();
	}

	/**
	 * 
	 * @param index
	 * @return the nth element from the top of the activation record stack. 
	 * 0 means the topmost,
	 * 1 means the one below the topmost, ...
	 */
	public Activation getFrame(int n){
		return stack.get(stack.size() - n - 1);
	}

	public void clearTopMostFrame(){
		peekTopMostFrame().clear();
	}

	/**
	 * @return the uncaughtExceptionInUninstrumented
	 */
	public boolean isUncaughtExceptionInUninstrumented() {
		return this.uncaughtExceptionInUninstrumented;
	}

	/**
	 */
	public void setUncaughtExceptionInUninstrumented() {
		this.uncaughtExceptionInUninstrumented = true;
	}

	/**
	 * @return the exceptionThrown
	 */
	public boolean isExceptionThrown() {
		return exceptionThrown;
	}

	/**
	 * @return the throwInstructionOccured
	 */
	public boolean isThrowInstructionOccured() {
		return throwInstructionOccured;
	}

	/**
	 * Indicate that an instance of ATHROW bytecode got executed.
	 */
	public void setThrowInstructionOccured(){
		this.throwInstructionOccured = true;
	}
	
	/**
	 */
	public void setExceptionThrown() {
		this.exceptionThrown = true;
	}
	
	public void setExceptionCaught(){
		this.exceptionThrown = false;
		this.uncaughtExceptionInUninstrumented = false;		//We have to set this one to false too!
		this.throwInstructionOccured = false;
	}
	
	/**
	 * Find the memory location corresponding to the local variable at the specified index from the top-most
	 * stack frame and returns the last binding for it. This method should never return null, because it is 
	 * supposed that each local variable is defined (has a binding) before being used.
	 * 
	 * @param localIndex
	 * @return
	 */
	public Binding<LocalMemoryLocation> readFromLocal(int localIndex){
		return peekTopMostFrame().readFromLocal(localIndex);
	}
	
	public Binding<LocalMemoryLocation>[] readAllArguments(){
		Activation frame = peekTopMostFrame();
		int argsCount = frame.getArgsCount();
		Binding<LocalMemoryLocation>[] bindings = new Binding[argsCount];
		int currLocalIndex = 0;
		for (int i = 0; i < bindings.length; i++) {
			bindings[i] = readFromLocal(currLocalIndex);
			currLocalIndex = frame.getNextLocalIndex(currLocalIndex);
		}
		return bindings;
	}
	
	/**
	 * 
	 * Writing into a local variable at bytecode-level means adding a new binding for either a currently
	 * defined local variable or a completely new local variable. In the former case, the new binding
	 * could also introduce a new local variable whose index overlaps with an already defined but died local var.
	 * We first check if there is already a binding for the given local variable and if the binding has the given
	 * type. If anyone of these two condition doesn't hold, we create a new location object and invalidate the 
	 * former one by overwriting it, otherwise we use the currently existing location. Finally, we change the 
	 * binding of the location to this new definition and overwrite this binding into the local variable array 
	 * of this frame.
	 * 
	 * Distinguishing between an already defined local variable from a local variable whose index is re-used 
	 * is not always possible. Consider the case when a local variable inside a block has the same type as 
	 * the local variable that is coming next out of this block with the same type. Even using the names for 
	 * comparison doesn't work because we can use a name that is already used within a block for defining another
	 * local variable outside after the block. This causes issues in counting the number of memory locations used
	 * but it doesn't have any effect on finding backward slices.
	 * 
	 * @param index
	 * @param type
	 * @return
	 */
	public Binding<LocalMemoryLocation> writeIntoLocal(int localIndex, int insnIndex, Value value, BytecodeEvent provider){
		return peekTopMostFrame().writeIntoLocal(localIndex, insnIndex, value, provider);
	}
	
	public Binding<StackSlotMemoryLocation> pushOnStack(Value value, Type type, BytecodeEvent provider){
		return peekTopMostFrame().pushOnStack(value, type, provider);
	}
	
	public Binding<StackSlotMemoryLocation> popOffStack(){
		return peekTopMostFrame().popStackSlotBinding();
	}
	
	public Binding<StackSlotMemoryLocation> popOffStackNoSideEffect(){
		return peekTopMostFrame().popStackSlotBindingNoSideEffect();
	}
	
	/**
	 * This method is used on entrance into a method, before pushing the 
	 * corresponding stack frame on thread call stack. 
	 */
	public boolean isCallerInstrumented(final String methodId, final int argsCount, boolean isStatic){
		final String methodName = Utils.findMethodName(methodId); 
		final String desc = Utils.findDescriptor(methodId);
		final String ownerClass = Utils.findClassName(methodId);
		
		if (stack.isEmpty())
			return false;
		
		Activation topMostFrame = peekTopMostFrame();
		//TODO: this condition is not sufficient for all cases like returning from an uninstrumented method into an instrumented, ... Should be updated
		if ((topMostFrame.getMethodName().equals(methodName) && topMostFrame.getDesc().equals(desc) && !topMostFrame.isInitialized())){
			return true;
		}
		
		return false;
	}

	public Activation addFrameFor(final String methodId, 
									final int argsCount, 
									final boolean isStatic, 
									final INVOKEARGS_Operation provider){
		
		final String methodName = Utils.findMethodName(methodId); 
		final String desc = Utils.findDescriptor(methodId);
		final String ownerClass = Utils.findClassName(methodId);
		
		
		if (stack.isEmpty()){	//This is the case for the main method 
			Activation frame = createAndPushFrame(methodName, desc, ownerClass, argsCount, isStatic, provider);    // should be updated in the next step with the information about locals and stack
			
			// The main method is called through reflection API, 
			// so we add this dummy pseudo variable for consistency 
			// with the rest of the system
			definePseudoVariable(BytecodeEvent.UNTRACED_BYTECODE_OPERATION, -1);
			return frame;
		}

		Activation frame = createAndPushFrame(methodName, desc, ownerClass, argsCount, isStatic, provider);   // should be updated in the next step with the information about locals and stack
		return frame;
	}
	
//	/**
//	 * This method checks if the frame below the top-most stack frame really belongs to the caller of
//	 * the top-most sack frame. This method assumes that the callee's frame is already pushed on the 
//	 * thread call stack.
//	 * 
//	 * TODO: the predicate for this checking is not complete and we should extend it such that it covers
//	 * all the cases where an uninstrumented method can call an instrumented method and vice-versa.
//	 * @return
//	 */
//	public boolean isCallerInstrumentedWithCallee(final String methodId, final int argsCount, boolean isStatic){
//		if (stack.size() == 0 || stack.size() == 1)
//			return false;
//		
//		final String methodName = Utils.findMethodName(methodId); 
//		final String desc = Utils.findDescriptor(methodId);
//		final String ownerClass = Utils.findClassName(methodId);
//		
//		Activation topmost = popFrame();
//		Activation caller = peekTopMostFrame();
//		pushFrame(topmost);
//		if ((caller.getMethodName().equals(methodName) && caller.getDesc().equals(desc) && !caller.isInitialized())){    
//			return true;
//		}
//		
//		return false;			
//	}
	
	public boolean isTopMostFrameInitialized(){
		return peekTopMostFrame().isInitialized();
	}
	
	public void addOperation(BytecodeEvent op){
		addOperation(op, true);
	}
	
	public void addOperation(BytecodeEvent op, boolean log){
		this.operations.add(op);
		
		Analyzer.log(op.toString(), log && Analyzer.DO_LOG);
	}
	
	/**
	 * Return the bytecode operation that reside {@code count} slots
	 * before the last operation.
	 * 0 means the last
	 * 1 means the one before the last
	 *  
	 * @param count
	 * @return
	 */
	public BytecodeEvent getBytecodeOperationBeforeLast(int count){
		if (this.operations.size() -1 - count >= 0)
			return this.operations.get(this.operations.size() - 1 - count);
		else
			return null;
	}
	
	public List<BytecodeEvent> getBytecodeOperations(){
		return this.operations;
	}

	/**
	 * This method is called only when the topmost frame belongs to a newly called method and its caller's frame
	 * is also already pushed on the stack.
	 * 
	 */
	public Binding<StackSlotMemoryLocation>[] useActualParameters(boolean callerInstrumented){
		if (!callerInstrumented){   //This happens for the first method executed (usually main) 
										//where there is no activation preceding it to provide its arguments				
			return Binding.UNDEFINED_BINDINGS;
		}
		
		Activation calleeFrame = popFrame();
		Activation callerFrame = peekTopMostFrame();
		int parametersCount = calleeFrame.getArgsCount();
		
		Binding<StackSlotMemoryLocation>[] params = new Binding[parametersCount];
		for (int i = parametersCount - 1; i >=0 ; i--) {
			params[i] = callerFrame.popStackSlotBinding();
		}
		pushFrame(calleeFrame);
		return params;
	}
	
	/**
	 * This method just finds the actual parameters on top of the caller's stack frame and returns a reference to
	 * them, without affecting the state of the stack, i.e., the arguments remain on top of the stack.
	 * 
	 * @return
	 */
	public Binding<StackSlotMemoryLocation>[] getActualParameters(boolean callerInstrumented){
		if (!callerInstrumented){				//This happens for the first method executed (usually main) where there is no activation preceding it to provide its arguments
			return Binding.UNDEFINED_BINDINGS;
		}
		
		Activation calleeFrame = popFrame();
		Activation callerFrame = peekTopMostFrame();
		int parametersCount = calleeFrame.getArgsCount();
		
		Binding<StackSlotMemoryLocation>[] params = callerFrame.getStackSlotBindings(parametersCount);
		
		pushFrame(calleeFrame);
		return params;
	}
	
	/**
	 * 
	 * @param values could have size 0 while there are some args. This happens for the main method where we don't have the origins of the arguments
	 * @return
	 */
	public Binding<LocalMemoryLocation>[] defActualParameters(Value[] values, BytecodeEvent provider){
		Activation frame = peekTopMostFrame();
		int parametersCount = frame.getArgsCount();
		Binding<LocalMemoryLocation>[] bindings = new Binding[parametersCount];
		int currLocalIndex = 0;
		
		for (int i=0; i<bindings.length; i++){
			Value value = values[i];
			bindings[i]= writeIntoLocal(currLocalIndex, 0, values.length == 0 ? Value.NOT_AVAILABLE: values[i], provider);
			currLocalIndex += value.getType().getCategory();
		}
		
		return bindings;
	}
	
	public Binding<ReturnValueMemoryLocation> useReturnValue(){
		return peekTopMostFrame().getLastInvokationFrame().useReturnValue();
	}
	
	public Binding<ReturnValueMemoryLocation> defReturnValue(Value value, BytecodeEvent provider){
		return peekTopMostFrame().defReturnValue(value, provider);
	}
	
	/**
	 * This method is called from INVOKERETURNING_Operation when we return from a method whose body was not
	 * instrumented. We call this method to artificially initialize its activation to show the flow of info
	 * from the args to the return value location.
	 * 
	 * @param classMethodDescString
	 * @param types
	 */
	public Activation initializeCalleeFrame(String classMethodDescString, Type[] types){
		Activation callee = peekTopMostFrame();
		callee.initialize(types, classMethodDescString);
		return callee;
	}
	
	public Type[] getArgTypes(){
		return peekTopMostFrame().getArgTypes();
	}
	
	public int getNextTempIndex(){
		return peekTopMostFrame().getNextTemporaryId();
	}
	
	public StackSnapshot takeStackSnapshot(){
		return peekTopMostFrame().takeSnapshot();
	}
	
	public int getPseudoVariableIndex(int basicBlockId){
		return bbVarLocations.indexOf(basicBlockId);
	}

	/**
	 * This method is used for defining a pseudo variable for
	 * NON-invocation control flow instructions. 
	 * 
	 * For invocation instructions see the overloaded version
	 * of this method.
	 *  
	 * @param provider
	 * @param location
	 * @param target
	 * @return
	 */
	public Binding<PseudoVariableMemoryLocation> definePseudoVariable(BytecodeEvent provider, PseudoVariableMemoryLocation location, int target){
		Binding<PseudoVariableMemoryLocation> binding = new Binding<PseudoVariableMemoryLocation>(location, new Value(target), provider); 
		basicBlocks.push(binding);
		bbVarLocations.add(location);
		return binding;
	}

	/**
	 * This method is used for defining a pseudo variable for
	 * an invocation instruction. Each invocation is assigned 
	 * a unique negative integer id. The combination of the
	 * invocation id and the basic block's local id within a
	 * method uniquely identifies a basic block globally during
	 * the execution.
	 * 
	 * @param provider
	 * @param value
	 * @return
	 */
	public Binding<PseudoVariableMemoryLocation> definePseudoVariable(BytecodeEvent provider, int value){
		Binding<PseudoVariableMemoryLocation> binding = definePseudoVariable(provider, new PseudoVariableMemoryLocation(peekTopMostFrame(), this.nextInvocationId--), value);
		return binding;
	}
	
	public int getAndIncrementNextInvocationId(){
		return this.nextInvocationId--;
	}
	
	@Override
	public List<? extends MemoryLocation> getElements() {
		List<PseudoVariableMemoryLocation> locations = new ArrayList<>();
		for (Binding<PseudoVariableMemoryLocation> binding : basicBlocks) {
			locations.add(binding.getMemoryLocation());
		}
		return locations;
	}

	@Override
	public int getElementCount() {
		return basicBlocks.size();
	}

	@Override
	public int getOffset() {
		return offset;
	}

	@Override
	public void setOffset(int offset) {
		this.offset = offset;
	}
	
	public Binding<PseudoVariableMemoryLocation> popOffBasicBlockSimulatorStack(){
		return this.basicBlocks.pop();
	}
	
	public Binding<PseudoVariableMemoryLocation> peekBasicBlockSimulatorStack(){
		return this.basicBlocks.peek();
	}
	
	public void pushOnBasicBlockSimulatorStack(Binding<PseudoVariableMemoryLocation> binding){
		this.basicBlocks.add(binding);
	}
	
	public void killInvoke(){
		Binding<PseudoVariableMemoryLocation> curr = null;
		do{} while ((curr = popOffBasicBlockSimulatorStack()).getMemoryLocation().getPseudoVariableId() >= 0);
		
		/*
		 * The curr variable will point to the pseudo variable created  
		 * as a result of the the last invocation up to this point.
		 * in case of PEIs, we need to keep the negative pseudo var
		 * on top of the basic block simulator stack, so that the 
		 * following basic block can use it and make itself control
		 * dependent on the invocation.
		 * 
		 * In case of the absence of PEIs, the instruction following
		 * an invocation is not control dependent on the invocation.
		 */
		if (Analyzer.includePEIs){
			pushOnBasicBlockSimulatorStack(curr);		
		}
	}
	
//	public void killInvokeAfterException(){
//		int currentInvocationId = peekTopMostFrame().getInvokationId();
//		Stack<Integer> temp = new Stack<>();
//		int pseudoVar;
//		while(true){
//			do{
//				pseudoVar = popOffBasicBlockSimulatorStack().getMemoryLocation().getPseudoVariableId();
//				temp.push(pseudoVar);
//			}while(pseudoVar >= 0);
//			if (pseudoVar == currentInvocationId){
//				while (!temp.isEmpty()){
////					basicBlocks
//				}
//			}
//		}
//	}

	/**
	 * @return the currentBasicBlockId
	 */
	public int getCurrentBasicBlockId() {
		return currentBasicBlockId;
	}

	/**
	 * @param currentBasicBlockId the currentBasicBlockId to set
	 */
	public void setCurrentBasicBlockId(int currentBasicBlockId) {
		this.currentBasicBlockId = currentBasicBlockId;
	}

	@Override
	public String toString() {
		String res = "";
		Iterator<Activation> it = stack.iterator();
		while (it.hasNext()){
			res += it.next().toString()+"\n";
		}
		return res;
	}

	public boolean isCallStackEmpty(){
		return stack.isEmpty();
	}
	
	public int getAndIncrementNextOperationIndex(){
		if (nextOperationIndex == Integer.MAX_VALUE)
			throw new IllegalStateException("The number of events overflowing...!");
		
		return nextOperationIndex++;
	}
	
	public Set<Activation> getActicationRecords(){
		return this.frames;
	}
	
	@Override
	public void accept(TraceElementVisitor visitor) throws VisitorException {
		visitor.visit(this);
	}

	/**
	 * @return the threadId
	 */
	public int getTId() {
		return threadId;
	}
	
	/**
	 * @return the analyze
	 */
	public boolean isTraceMode() {
		return this.traceMode;
	}

	/**
	 * @param analyze the analyze to set
	 */
	public void setTraceMode(boolean traceMode) {
		this.traceMode = traceMode;
	}

	@Override
	public Binding<? extends MemoryLocation>[] getElementsAtTime(long time) {
		//TODO implement the body 
		throw new NotImplementedException();
	}

	@Override
	public Binding<? extends MemoryLocation> getElementAtTime(int index, long time) {
		//TODO implement the body 
		throw new NotImplementedException();
	}

	@Override
	public Binding<? extends MemoryLocation> getElementAtTime(String name, long time) {
		//TODO implement the body 
		throw new NotImplementedException();
	}
	
	public BytecodeEvent getBytecodeEventAtIndex(int index){
		return operations.get(index); 
	}

	/**
	 * The solo purpose of this class is to support the situation
	 * when we simulate an event whose occurrence we didn't trace.
	 * The concrete example is an event of type InvokeArgs.
	 * When we enter into an instrumented method whose caller was
	 * not instrumened, we haven't tracked its invocation, but we
	 * still generate a dummy InvokeArgs operation for consistency
	 * purposes. This dummy InvokeArgs uses an instance of an 
	 * UntracedActivationRecord which in turn uses an instance of
	 * this class to represent a thread which we haven't tracked 
	 * at runtime. 
	 * 
	 * @author reza
	 *
	 */
	private static class UntracedThreadCallStack extends ThreadCallStack{
		public UntracedThreadCallStack(){
			super(-1);
		}
	}
}
