package ch.usi.inf.sape.tracer.analyzer.bytecodeops;

import ch.usi.inf.sape.tracer.analyzer.Activation;
import ch.usi.inf.sape.tracer.analyzer.Binding;
import ch.usi.inf.sape.tracer.analyzer.EventConstants;
import ch.usi.inf.sape.tracer.analyzer.ThreadCallStack;
import ch.usi.inf.sape.tracer.analyzer.TraceElementVisitor;
import ch.usi.inf.sape.tracer.analyzer.Value;
import ch.usi.inf.sape.tracer.analyzer.VisitorException;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.PseudoVariableMemoryLocation;

/**
 * UPDATE: A negative value for the arg bb means entry into
 * a new method. So, we should assign a new dynamically 
 * generated invocation id to the corresponding block.
 * 
 * Upon the entrance of the control flow into each basic 
 * block, an instance of this class will be generated and 
 * added to the list of bytecode events. 
 * 
 * A basic block is identified by a unique pseudo variable. 
 * A Pseudo variable is simply a unique integer id derived 
 * from the index of the leading bytecode in the corresponding 
 * basic block. 
 * 
 * The value of a pseudo variable memory location is the id
 * of the basic block from which the control flow flew into
 * it. This is used for the backward slicing purposes.
 *  
 * In order to track control dependencies, we need first 
 * need to know which basic block an instruction belongs
 * to and then find the control flow instruction that
 * dominates that basic block. The latter step can be
 * repeated recursively to find the whole control flow
 * dependency.
 * 
 * Finding the basic block to which an instruction belongs
 * is a linear traversal action on the set of bytecode 
 * events occurring before that instruction up to the point
 * where we encounter a basic block entry event. That event
 * indicates the enclosing basic block of the instruction.
 * 
 * The instruction that dominates a basic block is simply
 * the control flow instruction which causes the control flow
 * reach to that basic block. To find that, each control flow
 * instruction, in addition to its normal def values, has 
 * an extra definition for a pseudo variable. That pseudo
 * variable has the same id as its enclosing basic block, 
 * just as it is executed within its enclosing basic block,
 * and its value is also the value of its enclosing basic
 * block. The new binding for this value will be used in the
 * next basic block entry event to define the pseudo variable
 * corresponding to the newly entered basic block. This scenario
 * makes a connection between the control flow instruction and
 * its following entered basic block. In this way, we can always 
 * track the def-use chain to find the control flow dependencies.
 * An exception for this is the invocation instructions.
 * The id of basic blocks are unique within the body of their 
 * corresponding body, but not necessariy between different 
 * methods. To fix this issue and identify them basic blocks
 * uniquely during the whole exectuion, the pseudo variable
 * defined for an invocation instruction is treated differently.
 * An invocation instruction defines a pseduo variable that has
 * a negative integer id assigned dynamically at runtime and 
 * its value is just as before, the id of the basic block in
 * which the invocation happened.  The combination of the
 * invocation id and the basic block's local id within a
 * method uniquely identifies a basic block globally during
 * the execution.
 * 
 * @author reza
 *
 */
public class BasicBlockEntry_PseudoOperation extends BytecodeEvent{
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 6948581719688550694L;
	
	private final int bbId;
	
	public BasicBlockEntry_PseudoOperation(short opCode, int bytecodeIndex, int bb, String name, int[] doms){
		super(opCode, bytecodeIndex, findContainer());
		this.bbId = bb;
		ThreadCallStack tcs = ThreadCallStack.getCurrentThreadCallStack();
		
		if (!ThreadCallStack.getCurrentThreadCallStack().isTopMostFrameInitialized()){				//This is the case when we invoke an uninstrumented method
			ThreadCallStack.getCurrentThreadCallStack().popFrame();
		}
		
		// We need to find the basic block from which we enter into this basic block
		Binding<PseudoVariableMemoryLocation> use;
		
		//TODO remove constant -1 and use a constant variable instead
		if (bb == -1){		//This is the case for entering into a new method
			use = tcs.peekBasicBlockSimulatorStack();
		}
		else if (tcs.peekBasicBlockSimulatorStack().getMemoryLocation().getContainer() 
				!= tcs.peekTopMostFrame()){						//this means an uncaught exception in another method

			use = tcs.peekBasicBlockSimulatorStack();
			
			setExceptionFlag(tcs);
			
			//pop off the basic block stack simulator to the point we reach to a 
			//pseudo variable which belongs to the current topmost stack frame.
			removeExtraPseudoVariables(); 
		}else{
			while (true){
				int top = tcs.peekBasicBlockSimulatorStack().getMemoryLocation().getPseudoVariableId();
				if (canBeEnteredFrom(bb, top, doms)){
					use = tcs.peekBasicBlockSimulatorStack();
					break;
				}

				tcs.popOffBasicBlockSimulatorStack();
			}
		}
		
		this.uses = new Binding[]{use};
		PseudoVariableMemoryLocation location = new PseudoVariableMemoryLocation(tcs.peekTopMostFrame(), 
														bb < 0 ? tcs.getAndIncrementNextInvocationId() : bb);
		
		Binding<PseudoVariableMemoryLocation> binding = new Binding<>(location, 
				new Value(use.getMemoryLocation().getPseudoVariableId()), this);
		tcs.pushOnBasicBlockSimulatorStack(binding);
		
		this.defs = new Binding[]{binding};
	}

	public BasicBlockEntry_PseudoOperation(short opCode, int sourceLineNo,
			int opIndex, Activation activationRecord,
			Binding<? extends MemoryLocation>[] uses,
			Binding<? extends MemoryLocation>[] defs, int bbId) {
		super(opCode, sourceLineNo, opIndex, activationRecord, uses, defs);
		this.bbId = bbId;
	}

	@Override
	public Binding<PseudoVariableMemoryLocation>[] getUses() {
		return (Binding<PseudoVariableMemoryLocation>[])this.uses;
	}

	@Override
	public Binding<? extends MemoryLocation>[] getDefs() {
		return this.defs;
	}

	@Override
	public Binding<? extends MemoryLocation>[] getUsesForDef(
			Binding<? extends MemoryLocation> def) {
		return Binding.NOT_AVAILABLE_BINDINGS;
	}

	@Override
	public String getOperationName() {
		return "BASICBLOCK [entry]";
	}

	@Override
	public boolean isBasicBlockEntry() {
		return true;
	}

	/**
	 * Returns true if the current basic block can be entered from the top basic block
	 * 
	 * @param top
	 * @param dominators
	 * @return
	 */
	private boolean canBeEnteredFrom(int bb, int top, int[] dominators){
		if (bb < 0)
			return true;
		
		boolean hasNegativeDominator = false;
		for (int dominator : dominators) {
			if (dominator >= 0) {
				if (top == dominator){
					return true;
				}
			} else{
				hasNegativeDominator = true;
			}
		}
		
		if (hasNegativeDominator && top < 0)
			return true;
		
		return false;
	}
	
	private void removeExtraPseudoVariables(){
		ThreadCallStack tcs = ThreadCallStack.getCurrentThreadCallStack();
		Activation currentActivation = tcs.peekTopMostFrame();
		while (tcs.peekBasicBlockSimulatorStack().getMemoryLocation().getContainer() != currentActivation){
			tcs.popOffBasicBlockSimulatorStack();
		}
	}
	
	private static Activation findContainer(){
		ThreadCallStack tcs = ThreadCallStack.getCurrentThreadCallStack();
		Activation result;
		if (!tcs.isTopMostFrameInitialized()){				//This is the case when we invoke an uninstrumented method
			result = tcs.getFrame(1);
		}else{
			result = tcs.peekTopMostFrame();
		}

		return result;
	}
	
	private static void setExceptionFlag(ThreadCallStack tcs){
		if (!tcs.isExceptionThrown()){
			tcs.setUncaughtExceptionInUninstrumented();
		}else{
			tcs.setExceptionThrown();
		}
	}

	@Override
	public byte getEventClassCode() {
		return EventConstants.BASIC_BLOCK_ENTRY_EVENT;
	}
	
	@Override
	public void accept(TraceElementVisitor visitor) throws VisitorException {
		visitor.visit(this);
	}

}
