package ch.usi.inf.sape.tracer.analyzer.bytecodeops;

import ch.usi.inf.sape.tracer.analyzer.Activation;
import ch.usi.inf.sape.tracer.analyzer.Binding;
import ch.usi.inf.sape.tracer.analyzer.EventConstants;
import ch.usi.inf.sape.tracer.analyzer.ThreadCallStack;
import ch.usi.inf.sape.tracer.analyzer.TraceElementVisitor;
import ch.usi.inf.sape.tracer.analyzer.Value;
import ch.usi.inf.sape.tracer.analyzer.VisitorException;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocation;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocationInfo.Type;

/**
 * The catch block is connected to its corresponding (possibly existing) 
 * throw statement through the control flow dependency generated by 
 * BasicBlockEntry operation. That's why the use set of this operation is
 * set to no-use. 
 * @author reza
 *
 */
public class CatchBlock_Operation extends BytecodeEvent{

	/**
	 * 
	 */
	private static final long serialVersionUID = 8527387519980627025L;
	
	public CatchBlock_Operation(short opCode, int bytecodeIndex, Value exp) {
		super(opCode, bytecodeIndex, ThreadCallStack.getCurrentThreadCallStack().peekTopMostFrame());
		ThreadCallStack tcs = ThreadCallStack.getCurrentThreadCallStack();
		
		//After catching an exception, the stack frame is made empty 
		//and the exception object is pushed on top of stack
		tcs.clearTopMostFrame();
				
		if (tcs.isThrowInstructionOccured()){
			int i = 0;
			while (tcs.getBytecodeOperationBeforeLast(i).isBasicBlockEntry())		//find the athrow instruction
				i++;
			
			this.uses = new Binding[]{tcs.getBytecodeOperationBeforeLast(i).getDefs()[0]};
		}else{ 
				/*Either tcs.isUncaughtExceptionInUninstrumented() occurred or a PEI threw the exception. 
				 * There is no data dependency in this case, but just control dependency. The framework 
				 * can track control dependency automatically using the BasicBlockPseudoOps.
				 */
				this.uses = Binding.NO_USE_BINDINGS;
		}
			
		this.defs = new Binding[]{ThreadCallStack.getCurrentThreadCallStack().pushOnStack(exp, Type.REFERENCE, this)};
		
		tcs.setExceptionCaught();
	}
	
	public CatchBlock_Operation(short opCode, int sourceLineNo,
			int opIndex,
			Activation activationRecord,
			Binding<? extends MemoryLocation>[] uses,
			Binding<? extends MemoryLocation>[] defs) {
		super(opCode, sourceLineNo, opIndex, activationRecord, uses, defs);
	}

	@Override
	public Binding<? extends MemoryLocation>[] getUses() {
		return this.uses;
	}

	@Override
	public Binding<? extends MemoryLocation>[] getDefs() {
		return this.defs;
	}

	@Override
	public Binding<? extends MemoryLocation>[] getUsesForDef(Binding<? extends MemoryLocation> def) {
		if (def == null)
			return Binding.NOT_AVAILABLE_BINDINGS;
		
		if (def.equals(this.defs[0])){
			return this.uses;
		}
		
		return Binding.NOT_AVAILABLE_BINDINGS;
	}

	@Override
	public String getOperationName() {
		return "CATCH_BLOCK";
	}

	@Override
	public byte getEventClassCode() {
		return EventConstants.CATCH_BLOCK_EVENT;
	}

	@Override
	public void accept(TraceElementVisitor visitor) throws VisitorException {
		visitor.visit(this);
	}

}
