package ch.usi.inf.sape.tracer.analyzer.locations;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import ch.usi.inf.sape.tracer.analyzer.Binding;
import ch.usi.inf.sape.tracer.analyzer.TraceElementVisitor;
import ch.usi.inf.sape.tracer.analyzer.Utils;
import ch.usi.inf.sape.tracer.analyzer.Value;
import ch.usi.inf.sape.tracer.analyzer.VisitorException;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BytecodeEvent;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.MULTIANEWARRAY_Operation;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.XASTORE_Operation;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocationInfo.Type;

public class ArrayMetaData extends ObjectMetaData{

	private final Type elementType;
	
	private final List<Binding<ArrayElementMemoryLocation>>[] elementBindings;
	
	private final Binding<ArrayLengthMemoryLocation> lengthBinding;
	
	/**
	 * An array object can have different states within an execution.
	 * this field determines the time which the array currently
	 * is in. This is reflected in the values of bindings for the 
	 * array elements.
	 *  
	 */
	private long currentTime;
	
	private ArrayMetaData(ClassMetaData klass, int id, int size, Type elementType, BytecodeEvent provider) {
		super(klass, id, provider);
		this.elementType = elementType;
		this.elementBindings = new ArrayList[size];
		for (int i = 0; i < size; i++){
			this.elementBindings[i] = new ArrayList<>();
		}
		this.lengthBinding = new Binding<>(new ArrayLengthMemoryLocation(this), new Value(size), provider);
		this.currentTime = provider.getEventTime();
	}
	
	
	public static ArrayMetaData generateWithElementInit(ClassMetaData klass, int id, int size, Type elementType, boolean isTraced, BytecodeEvent provider) {
		ArrayMetaData array = new ArrayMetaData(klass, id, size, elementType, provider);
		
		for (int i = 0; i < size; i++) {
			ArrayElementMemoryLocation element = new ArrayElementMemoryLocation(array, i);
			array.initializeArrayElement(element, i, isTraced, provider);
		}
		
		return array;
	}
	
	public static ArrayMetaData generateWithoutElementInit(ClassMetaData klass, int id, int size, Type elementType, BytecodeEvent provider) {
		return new ArrayMetaData(klass, id, size, elementType, provider);
	}

//	/**
//	 * This constructor must only be used for an array that has subarrays as its elements
//	 * @param id
//	 * @param className
//	 * @param size
//	 * @param elementType
//	 * @param provider
//	 */
//	public ArrayMetaData(int id, ClassMetaData klass, int size, Type elementType, BytecodeEvent provider) {
//		super(klass, id, provider);
//		this.elementType = elementType;
//		this.elementBindings = new ArrayList[size];
//		this.lengthBinding = new Binding<>(new ArrayLengthMemoryLocation(this), new Value(size), provider);
//		this.currentTime = provider.getEventTime();
//	}

	
	@Override
	public ObjectType getObjectType() {
		return ObjectType.ARRAY;
	}
	
	public void initializeArrayElement(ArrayElementMemoryLocation location, int index, boolean isTraced, BytecodeEvent provider){
		Binding<ArrayElementMemoryLocation> binding;
		
		if (isTraced){
			binding = new Binding<>(location, elementType.getDefaultValue(), provider);
		}else{
			binding = new Binding<>(location, Value.NOT_AVAILABLE, BytecodeEvent.UNTRACED_BYTECODE_OPERATION);
		}
		
		this.elementBindings[index].add(binding);
		
		currentTime = provider.getEventTime();
	}
	
	public Binding<ArrayElementMemoryLocation>[] getElementBindings(){
		Binding<ArrayElementMemoryLocation>[] lastBindings = new Binding[this.elementBindings.length];
		for (int i = 0; i < lastBindings.length; i++) {
			lastBindings[i] = this.elementBindings[i].get(this.elementBindings[i].size() - 1);
		}
		return lastBindings;
	}
	
	/**
	 * @return the lengthBidning
	 */
	public Binding<ArrayLengthMemoryLocation> useLengthBinding() {
		return lengthBinding;
	}

	public Binding<ArrayElementMemoryLocation> useElementBinding(int index){
		return this.elementBindings[index].get(this.elementBindings[index].size() - 1);
	}
	
	public Binding<ArrayElementMemoryLocation> defElementBinding(int index, Value value, XASTORE_Operation provider){
		ArrayElementMemoryLocation elementLocation = this.elementBindings[index].get(this.elementBindings[index].size() - 1).getMemoryLocation();
		assert elementLocation != null;
		Binding<ArrayElementMemoryLocation> binding = new Binding<ArrayElementMemoryLocation>(elementLocation, value, provider);
		this.elementBindings[index].add(binding);
		this.currentTime = provider.getEventTime();
		return binding;
	}
	
	public Binding<ArrayElementMemoryLocation> defElementBinding(ArrayMetaData array, int index, Object subArray, MULTIANEWARRAY_Operation provider){
		Binding<ArrayElementMemoryLocation> binding = new Binding(new ArrayElementMemoryLocation(array, index), Utils.getValueForObject(subArray), provider);
		elementBindings[index].add(binding);
		this.currentTime = provider.getEventTime();
		return binding;
	}

	@Override
	public int getElementCount() {
		return elementBindings.length + 1;
	}
	
	@Override
	public List<MemoryLocation> getElements() {
		List<MemoryLocation> elementss = new ArrayList<>();
		for (List<Binding<ArrayElementMemoryLocation>> bindings : elementBindings) {
			elementss.add(bindings.get(0).getMemoryLocation());
		}
		elementss.add(lengthBinding.getMemoryLocation());
		return elementss;
	}

	@Override
	public void accept(TraceElementVisitor visitor) throws VisitorException {
		visitor.visit(this);
	}

	@Override
	public Binding<ArrayElementMemoryLocation>[] getElementsAtTime(long time) {
		Binding<ArrayElementMemoryLocation>[] result = new Binding[this.elementBindings.length];
		
		for (int i = 0; i < result.length; i++) {
			result[i] = (Binding<ArrayElementMemoryLocation>)getElementAtTime(i, time);
		}
		
		return result;
	}

	public int length(){
		return this.elementBindings.length;
	}


	@Override
	public Binding<? extends MemoryLocation> getElementAtTime(int elementIndex, long time) {
		List<Binding<ArrayElementMemoryLocation>> currentElementBindings = this.elementBindings[elementIndex];
		
		int counter = 0, index = 0;
		while (counter < currentElementBindings.size() && 
				time >= currentElementBindings.get(counter).getProvider().getEventTime()){
			index = counter;
			counter ++;
		}
		return this.elementBindings[elementIndex].get(index);
	}


	@Override
	public Binding<? extends MemoryLocation> getElementAtTime(String name, long time) {
		throw new UnsupportedOperationException();
	}
}
