package ch.usi.inf.sape.tracer.analyzer.locations;

import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.sun.tools.javac.jvm.Code;

import ch.usi.inf.sape.tracer.analyzer.Utils;
import ch.usi.inf.sape.tracer.analyzer.WeakIdentityHashMap2;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BytecodeEvent;
import ch.usi.inf.sape.tracer.analyzer.locations.MemoryLocationInfo.Type;

public class HeapSpace {
	
	/**
	 * The weak mapping from a Java object to the corresponding ObjectMemoryLocation instance
	 */
	private static WeakIdentityHashMap2<Object, ObjectMetaData> objMap = new WeakIdentityHashMap2<Object, ObjectMetaData>();

	/**
	 * Given that we use a weak identity hash map for keeping track of the 
	 * object to object location  mappings, we need a set to keep track of 
	 * all the object locations created during runtime
	 */
	private static final Set<InstanceMetaData> tracedInstanceMemoryLocations = new HashSet<>();
	
	private static final Set<ArrayMetaData> tracedArrayMemoryLocations = new HashSet<>();
	
	private static final Set<InstanceMetaData> untracedInstanceMemoryLocations = new HashSet<>();
	
	private static final Set<ArrayMetaData> untracedArrayMemoryLocations = new HashSet<>();
	
	/**
	 * Keeps track of the next id that can be used for an object location. 
	 */
	private static int nextInstanceId;
	
	private static int nextArrayId;
	
	private static List<HeapListener> listeners = new ArrayList<>();
	
	
	public static InstanceMetaData getOrDefineInstanceLocationMetaData(Object obj){
		InstanceMetaData objLoc = (InstanceMetaData) objMap.get(obj);
		if (objLoc == null){
			objLoc = defineInstanceLocationMetaData(obj, false, BytecodeEvent.UNTRACED_BYTECODE_OPERATION);
		}
		return objLoc;
	}
	
	/**
	 * Finds or creates an ArrayMemoryLocationMetaData instance. If not found, an instance
	 * will be created and considered as an untraced instance creation.
	 * 
	 * This method never returns null.
	 * 
	 * @param obj should not be null
	 * @return
	 */
	public static ArrayMetaData getOrDefineArrayLocationMetaData(Object obj){
		ArrayMetaData objLoc = (ArrayMetaData) objMap.get(obj);
		if (objLoc == null){
			objLoc = defineArrayLocationMetaData(obj, 
												Type.getTypeFor(obj.getClass().getComponentType().toString()), 
												false, 
												BytecodeEvent.UNTRACED_BYTECODE_OPERATION);
		}
		return objLoc;
	}
	
	/**
	 * 
	 * @param obj
	 * @param isTraced if the creation of this object is already traced. If not, we set the default field values to NA!
	 * @return
	 */
	public static InstanceMetaData defineInstanceLocationMetaData(Object obj, boolean isTraced, BytecodeEvent provider){
		InstanceMetaData instanceLocation = null;
		try{
			createInstanceBarrierBefore(obj, isTraced);
			Class klass = obj.getClass();  
			instanceLocation = new InstanceMetaData(GlobalSpace.getOrDefineClassLocationMetaData(klass), nextInstanceId ++, provider);
			objMap.put(obj, instanceLocation);
			createInstanceBarrier(obj, instanceLocation, isTraced);
			List<Field> fields = getAllFieldsInHierarchy(klass);
			for (Field field:fields){
				if(!Modifier.isStatic(field.getModifiers())){
					InstanceFieldMemoryLocation fieldLocation = new InstanceFieldMemoryLocation(instanceLocation, field.getName(), Utils.findTypeFromFullTypeName(field.getType().getName()));
					instanceLocation.initializeField(fieldLocation, isTraced, provider);
				}
			}
		}catch(Throwable t){  //We should never cause any interrupt in the normal program execution. In case the object is null, we postpone the exception to the point where the program really sees that value.
			t.printStackTrace(System.out);
			System.exit(0);
		}  
		
		return instanceLocation;
	}
	
	/**
	 * 
	 * @param obj
	 * @param type
	 * @param isTraced	if the creation of this array is already traced. If not, we set the default element values to NA!
	 * @return
	 */
	public static ArrayMetaData defineArrayLocationMetaData(Object obj, Type type, boolean isTraced, BytecodeEvent provider){
		ArrayMetaData arrayLocation = null;
		try{
			int arraySize = Array.getLength(obj);
			arrayLocation = ArrayMetaData.generateWithElementInit(GlobalSpace.getOrDefineClassLocationMetaData(obj.getClass()), nextArrayId ++, arraySize, type, isTraced, provider);
			objMap.put(obj, arrayLocation);
			createArrayBarrier(arrayLocation, isTraced);
			
		}catch(Throwable t){
			t.printStackTrace();	//We should never cause any interrupt in the normal program execution. In case the object is null, we postpone the exception to the point where the program really sees that value.
		}
		return arrayLocation;
	}
	
	/**
	 * This method is used just from MultiANewArray bytecode class to create subarrays.
	 * The difference with the normal array initialization is that the default value
	 * for subarrays is not null, but a pointer to the subarray in lower dimension.
	 * 
	 * @param obj
	 * @param type
	 * @param isTraced
	 * @param provider
	 * @return
	 */
	public static ArrayMetaData defineMultiArrayLocationMetaData(Object obj, Type type, boolean isTraced, BytecodeEvent provider){
		ArrayMetaData arrayLocation = null;
		try{
			int arraySize = Array.getLength(obj);
			arrayLocation = ArrayMetaData.generateWithoutElementInit(GlobalSpace.getOrDefineClassLocationMetaData(obj.getClass()), nextArrayId++, arraySize, type, provider);
			objMap.put(obj, arrayLocation);
			createArrayBarrier(arrayLocation, isTraced);
		}catch (Throwable t){
			t.printStackTrace();
		}
		
		return arrayLocation;
	}
	
//	public static ArrayMemoryLocation defineArrayLocationMetaData(Object obj){
//		
//	}
	
	private static List<Field> getAllFieldsInHierarchy(Class klass){
		List<Field> fields = new ArrayList<>(); 
		if (klass == null || klass == Object.class)
			return fields;
		
		fields.addAll(getAllFieldsInHierarchy(klass.getSuperclass()));
		fields.addAll(Arrays.asList(klass.getDeclaredFields()));
		return fields;
	}
	
	private static void createInstanceBarrierBefore(Object obj, boolean traced){

	}
	
	private static void createInstanceBarrier(Object obj, InstanceMetaData objLocation, boolean traced){
		for (HeapListener heapListener : listeners) {
			heapListener.addObjectEvent(obj, objLocation);
		}
		
		if (traced)
			tracedInstanceMemoryLocations.add(objLocation);
		else
			untracedInstanceMemoryLocations.add(objLocation);
	}
	
	private static void createArrayBarrier(ArrayMetaData arrayLocation, boolean traced){
		if (traced)
			tracedArrayMemoryLocations.add(arrayLocation);
		else 
			untracedArrayMemoryLocations.add(arrayLocation);
	}
	
	public static Set<InstanceMetaData> getAllocationTracedInstanceMetaData(){
		return tracedInstanceMemoryLocations;
	}
	
	public static Set<ArrayMetaData> getAllocationTracedArrayMetaData(){
		return tracedArrayMemoryLocations;
	}
	
	public static Set<InstanceMetaData> getAllocationNotTracedInstanceMetaData(){
		return untracedInstanceMemoryLocations;
	}
	
	public static Set<ArrayMetaData> getAllocationNotTracedArrayMetaData(){
		return untracedArrayMemoryLocations;
	}
	
	public static boolean contains(Object obj){
		return objMap.containsKey(obj);
	}
	
	public static ObjectMetaData getObjectLocationMetaData(Object obj){
		return objMap.get(obj);
	}
	
	public static Set<InstanceMetaData> getAllInstanceMetaData(){
		Set<InstanceMetaData> metaData = new HashSet<>(tracedInstanceMemoryLocations);
		metaData.addAll(untracedInstanceMemoryLocations);
		return metaData;
	}
	
	public static Set<Object> getAllStillAliveApplicationObjects(){
		return new HashSet<>(objMap.keySet());
	}
	
	/**
	 * Given a fully qualified class name, this method returns
	 * all meta data objects corresponding to each application
	 * object of the given type that existed during execution.
	 *  
	 * @param fullyQualifiedClassName e.g.: com.sun.tools.javac.jvm.Code
	 * @return
	 */
	public static Set<InstanceMetaData> getAllInstanceMetaDataForClass(String fullyQualifiedClassName){
		Set<InstanceMetaData> instances = new HashSet<>();
		for (InstanceMetaData instance : getAllInstanceMetaData()) {
			if (fullyQualifiedClassName.equals(instance.getClassName())){
				instances.add(instance);
			}
		}
		return instances;
	}
	
	public static void addHeapListener(HeapListener listener){
		listeners.add(listener);
	}
	
	public static List<HeapListener> getHeapListeners(){
		return new ArrayList<>(listeners);
	}
}
