package ch.usi.inf.sape.tracer.analyzer.locations;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ch.usi.inf.sape.tracer.analyzer.Binding;
import ch.usi.inf.sape.tracer.analyzer.TraceElementVisitor;
import ch.usi.inf.sape.tracer.analyzer.Value;
import ch.usi.inf.sape.tracer.analyzer.VisitorException;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.BytecodeEvent;
import ch.usi.inf.sape.tracer.analyzer.bytecodeops.PUTFIELD_Operation;

public class InstanceMetaData extends ObjectMetaData{
	
	/**
	 * keeps track of the last binding for each field based on the name
	 */
	private final Map<String, List<Binding<InstanceFieldMemoryLocation>>> lastBindings = new HashMap<>();
	
	/**
	 * used for presentation purposes.
	 */
	private int nextAvailableFieldOffset;
	
	/**
	 * An instance can have different states within an execution.
	 * this field determines the time which the instance currently
	 * is in. This is reflected in the values of bindings for the 
	 * fields.
	 *  
	 */
	private long currentTime;
	
	/**
	 * An instance of this class corresponds to the execution of a "new" bytecode. No fields get initialized (no binding), so
	 * one has to call the initialize() method right after creating an instance of this class. The body of the initialize method
	 * could have been embedded into this constructor, but in order to avoid passing an uninitialized "this" reference to some
	 * other method, we seperated these methods.
	 * 
	 * @param fieldsCount
	 */
	public InstanceMetaData(ClassMetaData klass, int id, BytecodeEvent provider) {
		super(klass, id, provider);
		this.currentTime = provider.getEventTime();
	}
	
	/**
	 * All fields are initialized to their default value upon the creation of an object. This method corresponds
	 * to the internal behavior of JVM when there is no constructor or some fields have missing explicit initializations.
	 * We could have done this within the body of the constructor, but in that case, we would have to pass uninitialized 
	 * "this" variable to the constructor of the InstanceFieldMemoryLocation that is risky.
	 * 
	 * @param fieldNames
	 * @param fieldTypes
	 */
	public void initializeField(InstanceFieldMemoryLocation fieldLoc, boolean isTraced, BytecodeEvent provider){
		Binding<InstanceFieldMemoryLocation> binding;
		if (isTraced){
			binding = new Binding<>(fieldLoc, fieldLoc.getType().getDefaultValue(), provider);
		}else{
			binding = new Binding<>(fieldLoc, Value.NOT_AVAILABLE, BytecodeEvent.UNTRACED_BYTECODE_OPERATION);
		}
		
		List<Binding<InstanceFieldMemoryLocation>> bindingsList = new ArrayList<>();
		bindingsList.add(binding);
		
		this.lastBindings.put(fieldLoc.getName(), bindingsList);
		this.currentTime = provider.getEventTime();
	}
	
	
	@SuppressWarnings("unchecked")
	public Binding<InstanceFieldMemoryLocation>[] getInstanceFieldBindings(){
		Binding<InstanceFieldMemoryLocation>[] bindings = new Binding[this.lastBindings.keySet().size()];
		int i = 0;
		for (String key : this.lastBindings.keySet()){
			List<Binding<InstanceFieldMemoryLocation>> bindings4key = this.lastBindings.get(key);
			bindings[i] = bindings4key.get(bindings4key.size() - 1);
			i++;
		}
				
		return bindings;
	}
	
	public Binding<InstanceFieldMemoryLocation> getInstanceFieldBinding(String fieldName){
		assert this.lastBindings.get(fieldName) != null;
		List<Binding<InstanceFieldMemoryLocation>> bindings4field = this.lastBindings.get(fieldName);
		return bindings4field.get(bindings4field.size() - 1);
	}
	
	
	public Binding<InstanceFieldMemoryLocation> defInstanceFieldBinding(String fieldName, Value value, PUTFIELD_Operation provider){
		List<Binding<InstanceFieldMemoryLocation>> bindings4field = this.lastBindings.get(fieldName);
		
		InstanceFieldMemoryLocation loc = bindings4field.get(0).getMemoryLocation();		//The memory location is same for all bindings 
		Binding<InstanceFieldMemoryLocation> newBinding = new Binding<>(loc, value, provider);
		bindings4field.add(newBinding);
		
		this.currentTime = provider.getEventTime();
		
		return newBinding;
	}

	
	@Override
	public ObjectType getObjectType() {
		return ObjectType.INSTANCE;
	}

	@Override
	public int getElementCount() {
		return lastBindings.keySet().size();
	}
	
	public int getAndIncrementNextAvailableFieldOffset(){
		return this.nextAvailableFieldOffset ++;
	}

	@Override
	public List<? extends MemoryLocation> getElements() {
		List<InstanceFieldMemoryLocation> fields = new ArrayList<>();
		Set<String> keys = lastBindings.keySet();
		for(String key : keys){
			fields.add(lastBindings.get(key).get(0).getMemoryLocation());		//The memory location is the same for all bindings
		}
		
		Collections.sort(fields, new Comparator<InstanceFieldMemoryLocation>() {

			@Override
			public int compare(InstanceFieldMemoryLocation o1, InstanceFieldMemoryLocation o2) {
				if (o1.getPresentationOffset() < o2.getPresentationOffset())
					return -1;
				else if (o1.getPresentationOffset() == o2.getPresentationOffset())
					return 0;
				
				return 1;
			}
		});
		
		return fields;
	}

	@Override
	public void accept(TraceElementVisitor visitor) throws VisitorException {
		visitor.visit(this);
	}

	@Override
	@SuppressWarnings("unchecked")
	public Binding<InstanceFieldMemoryLocation>[] getElementsAtTime(long time) {
		Binding<InstanceFieldMemoryLocation>[] result = new Binding[this.lastBindings.keySet().size()];
		
		int i = 0;
		for (String key : lastBindings.keySet()) {
			result[i] = (Binding<InstanceFieldMemoryLocation>) getElementAtTime(key, time);
			i++;
		}
		return result;
	}

	@Override
	public Binding<? extends MemoryLocation> getElementAtTime(int index, long time) {
		throw new UnsupportedOperationException();
	}

	@Override
	public Binding<? extends MemoryLocation> getElementAtTime(String name, long time) {
		List<Binding<InstanceFieldMemoryLocation>> currentFieldBindings = this.lastBindings.get(name);
		
		int counter = 0, index = 0;
		while (counter < currentFieldBindings.size() && 
				time >= currentFieldBindings.get(counter).getProvider().getEventTime()){
			index = counter;
			counter ++;
		}
		return currentFieldBindings.get(index);
	}

	/**
	 * TODO: This is a temporary method. Add this method to
	 * an appropriate interface
	 */
	public List<Binding<InstanceFieldMemoryLocation>> getAllBindingsFor(String name){
		return this.lastBindings.get(name);
	}
}
