import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import org.junit.rules.TestName;

import ch.usi.inf.sape.tracer.agent.UninstrumentableClassLoader;
import ch.usi.inf.sape.tracer.analyzer.BytecodeAnalyzer;
import ch.usi.inf.sape.tracer.analyzer.InfectoAnalyzer;
import ch.usi.inf.sape.tracer.analyzer.JUnitTestDriver;
import ch.usi.inf.sape.tracer.analyzer.RuntimeEnv;
import ch.usi.inf.sape.tracer.analyzer.ThreadCallStack;
import ch.usi.inf.sape.tracer.analyzer.locations.HeapListener;
import ch.usi.inf.sape.tracer.analyzer.locations.HeapSpace;


public class Driver {
	
	private enum RUNNING_MODE {TEST, INTERNAL, EXTERNAL, INFECTO};
	
	
	private static String appAbsPath = "";
	private static String appMainClass = "";
	private static String analyzerClassName = "";
	private static RUNNING_MODE runningMode = RUNNING_MODE.INTERNAL;
	private static String heapListenerClassName = "";
	private static URLClassLoader appCl;
	private final static List<URL> cpElements = new ArrayList<>();
	private static String testClassName = "";
	private static String testMethodName = "";

	public static void main(String[] args) {
		
		// Find the values for application absolute path
		// application main class, analyzer class name,
		// and classpath elements.
		parseArgs(args);
		
		URLClassLoader cl = null;
		try {
			cl = new UninstrumentableClassLoader(cpElements.toArray(new URL[0]));
			
			if (!heapListenerClassName.equals("")){
				Class<?> heapListenerClass = cl.loadClass(heapListenerClassName);
				HeapSpace.addHeapListener((HeapListener)heapListenerClass.newInstance());
			}
			
			appCl = new URLClassLoader(cpElements.toArray(new URL[0]));
			if (runningMode == RUNNING_MODE.INFECTO){
				JUnitTestDriver driver = new JUnitTestDriver(testClassName, testMethodName, appCl);
				driver.launch();
			}else{
				Class<?> appClass = appCl.loadClass(appMainClass);
				Method mainMethod = appClass.getMethod("main", new Class[]{String[].class});
				mainMethod.invoke(null, new Object[]{new String[0]});
			}
		}catch(Exception e){
			e.printStackTrace();
		}
		
		if (analyzerClassName.isEmpty())
			System.exit(0);
		
		try{	
			RuntimeEnv env = RuntimeEnv.getInstance();
			performPostExecutionActions(env);
			
			//URLClassLoader cl = new UninstrumentableClassLoader(((URLClassLoader)Thread.currentThread().getContextClassLoader()).getURLs(), null);
			Class<?> analyzerClass = cl.loadClass(analyzerClassName);
			Method analyzerMethod = analyzerClass.getMethod("analyze", new Class[]{RuntimeEnv.class});
			BytecodeAnalyzer bytecodeAnalyzer = (BytecodeAnalyzer) analyzerClass.newInstance();
			analyzerMethod.invoke(bytecodeAnalyzer, env);
			cl.close();
		} catch (ClassNotFoundException | 
				NoSuchMethodException | 
				SecurityException | 
				IllegalAccessException | 
				IllegalArgumentException | 
				InvocationTargetException | 
				InstantiationException | 
				IOException e) {
			e.printStackTrace();
		}
		
	}
	
	public static void parseArgs(String args[]){
		try{
			if (args.length == 0){
				throw new IllegalArgumentException("Zero arguments ...");
			}
			
			String appCP = "";
			for(String arg:args){
				if (arg.startsWith("-Dapp=")){
					appAbsPath += arg.substring(6);
				} else if (arg.startsWith("-Danalyzer=")){
					analyzerClassName += arg.substring(11);
				} else if (arg.startsWith("-Dcp=")){
					appCP += arg.substring(5);
				} else if (arg.startsWith("-Dmain=")){
					appMainClass += arg.substring(7);
				} else if (arg.startsWith("-Dhl=")){
					heapListenerClassName += arg.substring(5);
				} else if (arg.startsWith("-Dmode=")){
					String modeString = arg.substring(7);
					runningMode = RUNNING_MODE.valueOf(modeString);
				} else if (arg.startsWith("-Dtc=")){
					testClassName += arg.substring(5);
				} else if (arg.startsWith("-Dtm=")){
					testMethodName += arg.substring(5);
				}
			}

			switch (runningMode){
			case INTERNAL:
			{
				if (appMainClass.trim().equals("")){
					throw new IllegalArgumentException("The main class is missing ...");
				}
				
				break;
			}
			case EXTERNAL:
			{
				if (!appAbsPath.startsWith("/")){
					throw new IllegalArgumentException("Application path must be an absolute path ...");
				}
				
				if (appMainClass.trim().equals("")){
					throw new IllegalArgumentException("The main class is missing ...");
				}
				break;
			}
			case INFECTO:
			{
				analyzerClassName = "ch.usi.inf.sape.tracer.analyzer.InfectoAnalyzer";
				
				if (!appAbsPath.startsWith("/")){
					throw new IllegalArgumentException("Application path must be an absolute path ...");
				}
				
				break;
			}
			case TEST:
			{
				break;
			}
			}

			final StringTokenizer tokenizer = new StringTokenizer(appCP,":");
			while (tokenizer.hasMoreTokens()){
				String cpElement = tokenizer.nextToken();
				if (cpElement.startsWith("/"))
					cpElements.add(new URL("file://"+cpElement));
				else
					cpElements.add(new URL("file://"+appAbsPath + "/"+ cpElement));
			}

			for (URL url :((URLClassLoader)Thread.currentThread().getContextClassLoader()).getURLs()){
				cpElements.add(url);
			}

		}catch(IllegalArgumentException | MalformedURLException e){
			System.err.println("\nUsage: java Driver [-Dapp=<Absolute path of application>] [-Dcp=<Semicolon-sepearted list of application classpath elements>] [-Danalyzer=<analyzer class>]  -Dmain=<Main class>");
			e.printStackTrace(System.err);
			System.exit(-1);
		}
	}
	
	/**
	 * Any action needed to be done right after the application
	 * execution is done and before the analysis starts must
	 * happen in this method. 
	 * 
	 * Actions done:
	 * 1- Set all threads to non-trace mode
	 * 2- Store the application class loader to be used by analyzer
	 * 
	 * @param env
	 */
	public static void performPostExecutionActions(RuntimeEnv env){
		System.out.println("Application Exit!");
		//Set all threads to non-trace mode
		for (ThreadCallStack tcs : env.getThreadCallStacks()){
			tcs.setTraceMode(false);
		}
		
		//Store the application class loader to be used by analyzer
		env.setContextClassLoader(appCl);
		env.setAppAbsPath(appAbsPath);
	}

	
}

